import sys

import autograder.api.constants
import autograder.api.user.add
import autograder.cli.common
import autograder.util.hash

def run(arguments):
    arguments = vars(arguments)
    arguments['new-users'] = _load_users(arguments['path'])

    result = autograder.api.user.add.send(arguments, exit_on_error = True)

    autograder.cli.common.list_add_users(result, table = arguments['table'])
    return 0

def _load_users(path):
    users = []

    with open(path, 'r') as file:
        lineno = 0
        for line in file:
            lineno += 1

            line = line.strip()
            if (line == ""):
                continue

            parts = line.split("\t")
            if (len(parts) > 5):
                raise ValueError(
                    "File ('%s') line (%d) has too many values. Max is 5, found %d." % (
                        path, lineno, len(parts)))

            email = parts[0]

            password = ''
            if (len(parts) >= 2):
                password = parts[1]
                if (password != ''):
                    password = autograder.util.hash.sha256_hex(password)

            name = ''
            if (len(parts) >= 3):
                name = parts[2]

            role = 'unknown'
            if (len(parts) >= 4):
                role = parts[3]

            if (role not in autograder.api.constants.ROLES):
                raise ValueError(
                    "File ('%s') line (%d) has an invalid role '%s'." % (
                        path, lineno, role))

            lms_id = ''
            if (len(parts) >= 5):
                lms_id = parts[4]

            users.append({
                'email': email,
                'pass': password,
                'name': name,
                'role': role,
                'lms-id': lms_id,
            })

    return users

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.user.add._get_parser()

    parser.description = ('Add users to the course.'
            + ' When force is true, this becomes an upsert (update if exists, otherwise insert).')

    parser.add_argument('path', metavar = 'PATH',
        action = 'store', type = str,
        help = 'Path to a TSV file where each line contains up to five columns:'
                + ' [email, pass, name, role, lms-id].'
                + ' Only the email is required.')

    parser.add_argument('--table', dest = 'table',
        action = 'store_true', default = False,
        help = 'Output the results as a TSV table with a header (default: %(default)s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import autograder.api.submission.fetchscores

HEADER = [
    'email', 'has_submission', 'short-id', 'score', 'grading_start_time', 'message',
]

def run(arguments):
    result = autograder.api.submission.fetchscores.send(arguments, exit_on_error = True)

    print("\t".join(HEADER))

    for (email, submission) in sorted(result['submission-infos'].items()):
        row = [email, (submission is not None)]

        if (submission is None):
            row += ([''] * 4)
        else:
            row += [submission[key] for key in HEADER[2:]]

        print("\t".join(map(str, row)))

    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.submission.fetchscores._get_parser()
    return parser

if (__name__ == '__main__'):
    sys.exit(main())



"""
Commands for administering an autograder course/server.
"""

import os
import sys

import autograder.util.cli

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
ROOT_DIR = os.path.join(THIS_DIR, '..', '..', '..')

def run():
    relpath = os.path.relpath(THIS_DIR, start = ROOT_DIR)
    package = '.'.join(relpath.split(os.sep))

    print(__doc__.strip())
    autograder.util.cli.list_dir(THIS_DIR, package)
    return 0

def main():
    return run()

if (__name__ == '__main__'):
    sys.exit(main())



HEADERS = ['email', 'name', 'role', 'lms-id']
SYNC_HEADERS = HEADERS + ['operation']

SYNC_USERS_KEYS = [
    ('add-users', 'Added', 'add'),
    ('mod-users', 'Modified', 'mod'),
    ('del-users', 'Deleted', 'delete'),
    ('skip-users', 'Skipped', 'skip'),
]

def list_users(users, table = False):
    if (table):
        _list_users_table(users)
    else:
        _list_users(users)

def _list_users(users, indent = ''):
    for user in users:
        print(indent + "Email:", user['email'])
        print(indent + "Name:", user['name'])
        print(indent + "Role:", user['role'])
        print(indent + "LMS ID:", user['lms-id'])
        print()

def _list_users_table(users, header = True, keys = HEADERS):
    if (header):
        print("\t".join(keys))

    for user in users:
        row = [user[key] for key in keys]
        print("\t".join([str(value) for value in row]))

def list_sync_users(sync_users, table = False):
    if (table):
        _list_sync_users_table(sync_users)
    else:
        _list_sync_users(sync_users)

def _list_sync_users(sync_users):
    count = (len(sync_users['add-users'])
            + len(sync_users['mod-users'])
            + len(sync_users['del-users']))
    print("Synced %d users." % (count))

    for (key, label, _) in SYNC_USERS_KEYS:
        users = sync_users[key]
        if (len(users) == 0):
            continue

        print("%s Users:" % (label))
        _list_users(users, indent = '    ')

def _list_sync_users_table(sync_users):
    print("\t".join(SYNC_HEADERS))

    for (key, _, op) in SYNC_USERS_KEYS:
        users = sync_users[key]
        for user in users:
            user['operation'] = op

        _list_users_table(users, header = False, keys = SYNC_HEADERS)

def list_add_users(result, table = False):
    errors = result['errors']
    if ((errors is not None) and (len(errors) > 0)):
        print("Encounted %d errors." % (len(errors)))
        for error in errors:
            print("    Index: %d, Email: '%s', Message: '%s'." % (
                error['index'], error['email'], error['message']))

    list_sync_users(result, table = table)



"""
Look in a package for modules that look like CLI tools
(has a _get_parser() method that returns an argparse parser)
and list them.
"""

import os

import autograder.code

def handle_file(path, cmd):
    try:
        module = autograder.code.import_path(path)
    except Exception:
        return

    if ('_get_parser' not in dir(module)):
        return

    parser = module._get_parser()
    parser.prog = 'python3 -m ' + cmd

    print()
    print(cmd)
    print(parser.description)
    parser.print_usage()

def handle_dir(path, cmd):
    try:
        module = autograder.code.import_path(os.path.join(path, '__main__.py'))
    except Exception:
        return

    description = module.__doc__.strip()

    print()
    print(cmd + '.*')
    print(description)
    print("See `python3 -m %s` for more information." % (cmd))

def list_dir(base_dir, command_prefix):
    for dirent in sorted(os.listdir(base_dir)):
        path = os.path.join(base_dir, dirent)
        cmd = command_prefix + '.' + os.path.splitext(dirent)[0]

        if (dirent.startswith('__')):
            continue

        if (os.path.isfile(path)):
            handle_file(path, cmd)
        else:
            handle_dir(path, cmd)



"""
Handle interfacing with git repos.
"""

import os

import git

def ensure_repo(url, path, update = False, ref = None):
    if (os.path.isfile(path)):
        raise ValueError("Target git path exists and is a file: '%s'." % (path))

    if (not os.path.exists(path)):
        clone(url, path)

    repo = get_repo(path)

    if (update):
        update_repo(repo)

    if (ref is not None):
        checkout_repo(repo, ref)

def get_repo(path):
    return git.Repo(path)

def clone(url, path):
    return git.Repo.clone_from(url, path)

def checkout_repo(repo, ref):
    repo.git.checkout(ref)

def update_repo(repo):
    fetch_results = repo.remotes.origin.pull()

    for fetch_result in fetch_results:
        if (fetch_result.ref.name != ("origin/%s" % (repo.active_branch.name))):
            continue

        if (fetch_result.flags == git.remote.FetchInfo.HEAD_UPTODATE):
            return False
        else:
            return True



import argparse
import sys

import autograder.code

def run(args):
    text = autograder.code.extract_code(args.path)
    ast = autograder.code.sanitize_code(text)
    source = autograder.code.ast_to_source(ast)

    if (args.out_path is None):
        print(source)
        return 0

    with open(args.out_path, 'w') as file:
        file.write(source + "\n")

def _get_parser():
    parser = argparse.ArgumentParser(description =
        "Pull the code from a file and output the sanatized version as if it was being graded.")

    parser.add_argument('path', metavar = 'PATH',
        action = 'store', type = str,
        help = 'Path to the file to extract code from.')

    parser.add_argument('-o', '--out-path', dest = 'out_path',
        action = 'store', type = str, default = None,
        help = 'If specified, the code will be written here instead of output on stdout.')

    return parser

def main():
    return run(_get_parser().parse_args())

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import autograder.api.submission.fetchsubmission
import autograder.cli.submission.common

def run(arguments):
    result = autograder.api.submission.fetchsubmission.send(arguments, exit_on_error = True)

    if (not result['found-user']):
        print("No matching user found.")
        return 1

    if (not result['found-submission']):
        print("No matching submission found.")
        return 2

    autograder.cli.submission.common.output_grading_result(
        result['grading-result'], arguments.out_dir)

    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.submission.fetchsubmission._get_parser()

    parser.add_argument('-o', '--out-dir', dest = 'out_dir',
        action = 'store', type = str, default = '.',
        help = ('Where to create a new directory that contains the submission information.'
            + ' An existing subdirectory will be removed.'
            + ' Defaults to the current directory.'))

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import autograder.api.common
import autograder.api.config

API_ENDPOINT = 'user/add'
API_PARAMS = [
    autograder.api.config.PARAM_COURSE_ID,
    autograder.api.config.PARAM_USER_EMAIL,
    autograder.api.config.PARAM_USER_PASS,

    autograder.api.config.PARAM_FORCE,
    autograder.api.config.PARAM_DRY_RUN,
    autograder.api.config.PARAM_SKIP_EMAILS,
    autograder.api.config.PARAM_SKIP_LMS_SYNC,

    autograder.api.config.APIParam('new-users',
        'A list of users to add or modify.',
        required = True, cli_param = False),
]

DESCRIPTION = ('Add one or more users to the course.'
        + ' When force is true, this becomes an upsert (update if exists, otherwise insert).')

def send(arguments, **kwargs):
    return autograder.api.common.handle_api_request(arguments, API_PARAMS, API_ENDPOINT, **kwargs)

def _get_parser():
    parser = autograder.api.config.get_argument_parser(
        description = DESCRIPTION,
        params = API_PARAMS)

    return parser



import argparse
import sys

import autograder.style

def run(args):
    count, total_lines = autograder.style.check_paths(args.paths)
    print("Found %d style errors." % (count))

    if (count > 0):
        for (path, lines) in total_lines:
            print("\nStyle Errors for '%s':" % (path))
            print('---')
            print("\n".join(lines))
            print("---\n")

    return count

def _get_parser():
    parser = argparse.ArgumentParser(description =
        "Check the style of all '.py' and '.ipynb' files in the paths specificed (recursivley).")

    parser.add_argument('paths', metavar = 'path',
        type = str, nargs = '+',
        help = 'A path to check for style.')

    return parser

def main():
    return run(_get_parser().parse_args())

if (__name__ == '__main__'):
    sys.exit(main())



import argparse
import json
import os
import sys

import platformdirs

import autograder.api.constants
import autograder.api.error
import autograder.util.hash

DEFAULT_CONFIG_FILENAME = 'config.json'
DEFAULT_USER_CONFIG_PATH = platformdirs.user_config_dir('autograder.json')

CONFIG_PATHS_KEY = 'config_paths'

class APIParam(object):
    def __init__(self, key, description,
            config_key = None, required = True, cli_param = True,
            parser_options = {'action': 'store', 'type': str},
            hash = False):
        self.key = str(key)
        if ((key is None) or (self.key == '')):
            raise autograder.api.error.APIError("APIParam cannot have an empty key.")

        self.description = str(description)
        if ((description is None) or (self.description == '')):
            raise autograder.api.error.APIError("APIParam cannot have an empty description.")

        self.config_key = config_key
        if (self.config_key is None):
            self.config_key = key

        self.required = required
        self.cli_param = cli_param
        self.parser_options = parser_options
        self.hash = hash

def parse_api_config(config, params,
        additional_required_keys = ['server'],
        additional_optional_keys = ['verbose'],
        exit_on_error = False):
    """
    Given a tiered config and api parameters,
    return a dict that can be directly sierialized and sent to the autograder.
    Any hashed params that are not empty will be hashed.
    If |exit_on_error| is true sys.exit() will be called on an error,
    otherwise an error will be raised on an error.
    Any keys in |additional_*_keys| will be returned in a second dict.

    Returns: (<values derived from params>, <additional values>)
    """

    try:
        return _parse_api_config(config, params,
                additional_required_keys, additional_optional_keys)
    except autograder.api.error.APIError as ex:
        if (exit_on_error):
            print("ERROR: " + ex.args[0], file = sys.stderr)
            sys.exit(1)

        raise ex

def _parse_api_config(config, params, additional_required_keys, additional_optional_keys):
    data = {}
    extra = {}

    for param in params:
        if (param.config_key not in config):
            if (param.required):
                raise autograder.api.error.APIError(
                    f"Required parameter '{param.config_key}' not set.")

            continue

        value = config[param.config_key]
        if (param.hash and (value != '')):
            value = autograder.util.hash.sha256_hex(value)

        data[param.key] = value

    for key in additional_required_keys:
        if (key not in config):
            raise autograder.api.error.APIError(f"Required parameter '{key}' not set.")

        extra[key] = config[key]

    for key in additional_optional_keys:
        if (key not in config):
            continue

        extra[key] = config[key]

    return data, extra

def get_tiered_config(cli_arguments, skip_keys = [CONFIG_PATHS_KEY], show_sources = False):
    """
    Get all the tiered configuration options (from files and CLI).
    If |show_sources| is True, then an addition dict will be returned that shows each key,
    and where that key came from.
    """

    config = {}
    sources = {}

    if (isinstance(cli_arguments, argparse.Namespace)):
        cli_arguments = vars(cli_arguments)

    # Check the current directory config.
    if (os.path.isfile(DEFAULT_CONFIG_FILENAME)):
        with open(DEFAULT_CONFIG_FILENAME, 'r') as file:
            for key, value in json.load(file).items():
                config[key] = value
                sources[key] = "<default config file>::" + DEFAULT_CONFIG_FILENAME

    # Check the user config file.
    if (os.path.isfile(DEFAULT_USER_CONFIG_PATH)):
        with open(DEFAULT_USER_CONFIG_PATH, 'r') as file:
            for key, value in json.load(file).items():
                config[key] = value
                sources[key] = "<user config file>::" + DEFAULT_USER_CONFIG_PATH

    # Check the config files specified on the command-line.
    config_paths = cli_arguments.get(CONFIG_PATHS_KEY, [])
    if (config_paths is not None):
        for path in config_paths:
            with open(path, 'r') as file:
                for key, value in json.load(file).items():
                    config[key] = value
                    sources[key] = "<cli config file>::" + path

    # Finally, any command-line options.
    for (key, value) in cli_arguments.items():
        if (key in skip_keys):
            continue

        if ((value is None) or (value == '')):
            continue

        config[key] = value
        sources[key] = "<cli argument>"

    if (show_sources):
        return config, sources

    return config

def get_argument_parser(
        description = 'Send an API request to the autograder.',
        params = [],
        include_assignment = True):
    """
    Create an argparse parser that has all the standard options for API requests.
    """

    parser = argparse.ArgumentParser(description = description)

    parser.add_argument('--config', dest = CONFIG_PATHS_KEY,
        action = 'append', type = str,
        help = 'A JSON config file with your submission/authentication details.'
            + " Can be specified multiple times with later values overriding earlier ones."
            + " Config values can be specified in multiple places"
            + " (with later values overriding earlier values):"
            + " First './%s'," % (DEFAULT_CONFIG_FILENAME)
            + " then '%s'," % (DEFAULT_USER_CONFIG_PATH)
            + " now any files specified using --config in the order they were specified,"
            + " and finally any variables specified directly on the command line (like --user).")

    parser.add_argument('--server', dest = 'server',
        action = 'store', type = str, default = None,
        help = 'The URL of the server to submit to (default: %(default)s).')

    parser.add_argument('-v', '--verbose', dest = 'verbose',
        action = 'store_true', default = False,
        help = 'Output detailed information about the API request and response'
            + " (default: %(default)s).")

    for param in params:
        if (not param.cli_param):
            continue

        parser.add_argument(f'--{param.config_key}', dest = param.config_key,
            help = param.description,
            **param.parser_options)

    return parser

# Common API params.

PARAM_ASSIGNMENT_ID = APIParam('assignment-id',
        'The ID of the assignment to make this request to.',
        config_key = 'assignment', required = True)

PARAM_COURSE_ID = APIParam('course-id',
        'The ID of the course to make this request to.',
        config_key = 'course', required = True)

PARAM_DRY_RUN = APIParam('dry-run',
        'Do not commit/finalize the operation,'
            + ' just do all the steps and state what the result would look like.',
        required = False,
        parser_options = {'action': 'store_true', 'default': False})

PARAM_FILTER_ROLE = APIParam('filter-role',
        'Only show results from users with this role (all roles if unknown (default)).',
        required = False,
        parser_options = {'action': 'store', 'default': 'unknown',
            'choices': autograder.api.constants.ROLES})

PARAM_FORCE = APIParam('force',
        'Force the operation, overwriting and existing resources.',
        required = False,
        parser_options = {'action': 'store_true', 'default': False})

PARAM_NEW_PASS = APIParam('new-pass',
        'The new password to set for the user that is the target of this request.',
        required = False, hash = True)

PARAM_SKIP_EMAILS = APIParam('skip-emails',
        'Skip sending any emails. Be aware that this may result in inaccessible information.',
        required = False,
        parser_options = {'action': 'store_true', 'default': False})

PARAM_SKIP_LMS_SYNC = APIParam('skip-lms-sync',
        'Skip syncing with the LMS.',
        required = False,
        parser_options = {'action': 'store_true', 'default': False})

PARAM_TARGET_EMAIL = APIParam('target-email',
        'The email of the user that is the target of this request.',
        required = True)

PARAM_TARGET_EMAIL_OR_SELF = APIParam('target-email',
        'The email of the user that is the target of this request (defaults to you).',
        required = False)

PARAM_TARGET_PASS = APIParam('target-pass',
        'The password of the user that is the target of this request.',
        required = True, hash = True)

PARAM_TARGET_SUBMISSION_OR_RECENT = APIParam('target-submission',
        'The ID of the submission (default to the most recent submission).',
        required = False)

PARAM_USER_EMAIL = APIParam('user-email',
        'The email of the user making this request.',
        config_key = 'user', required = True)

PARAM_USER_PASS = APIParam('user-pass',
        'The password of the user making this request.',
        config_key = 'pass', required = True, hash = True)



import sys

import autograder.api.admin.coursereload

def run(arguments):
    autograder.api.admin.coursereload.send(arguments, exit_on_error = True)
    print("Course reloaded.")
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    return autograder.api.admin.coursereload._get_parser()

if (__name__ == '__main__'):
    sys.exit(main())



"""
Utilities for extracting and working with Python source code.
"""

import ast
import importlib.util
import json
import os
import sys
import types
import uuid

import autograder.utils

ALLOWED_EXTENSIONS = ['.py', '.ipynb']
AST_NODE_WHITELIST = [ast.Import, ast.ImportFrom, ast.FunctionDef, ast.ClassDef]

# TODO(eriq): Code from a python file should be cleaned using an AST as well.
def extract_code(path):
    """
    Gets the source code out of a path (to either a notebook or vanilla python).
    """

    code = None

    if (path.endswith('.ipynb')):
        code = extract_notebook_code(path)
    elif (path.endswith('.py')):
        with open(path, 'r') as file:
            lines = file.readlines()
        lines = [line.rstrip() for line in lines]

        code = "\n".join(lines) + "\n"
    else:
        raise ValueError("Unknown extension for extracting code: '%s'." % (path))

    return code.strip()

def extract_notebook_code(path):
    """
    Extract all the code cells from an iPython notebook.
    A concatenation of all the cells (with a newline between each cell) will be output.
    """

    with open(path, 'r') as file:
        notebook = json.load(file)

    contents = []

    for cell in notebook['cells']:
        if (cell['cell_type'] != 'code'):
            continue

        cell_code = ''.join(cell['source']).strip()

        # Ignore empty cells.
        if (cell_code == ''):
            continue

        contents.append(cell_code)

    return "\n".join(contents) + "\n"

def import_path(path, module_name = None):
    if (module_name is None):
        module_name = str(uuid.uuid4()).replace('-', '')

    # If it's a notebook, extract the code first and put it in a temp file.
    if (path.endswith('.ipynb')):
        source_code = extract_code(path)
        path = autograder.utils.get_temp_path(suffix = '.py')
        with open(path, 'w') as file:
            file.write(source_code)

    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module

def sanitize_and_import_path(path, syspath = None, **kwargs):
    """
    Get the code from a source file, sanitize it, exec it, and return it as a namespace (module).
    Sanitization in this context means removing things that are not
    imports, functions, constants, and classes.
    A "constant" will be considered an assignment where the LHS is a single variable all in caps.
    """

    if (syspath is None):
        syspath = os.path.dirname(os.path.abspath(path))

    source_code = extract_code(path)

    return sanitize_and_import_code(source_code, path, syspath = syspath, **kwargs)

def sanitize_and_import_code(source_code, path, as_dict = False, syspath = None):
    """
    See sanitize_and_import_path().
    """

    if (syspath is None):
        syspath = os.path.dirname(os.path.abspath(os.getcwd()))

    module_ast = sanitize_code(source_code)

    globals_defs = {}

    try:
        sys.path.append(syspath)
        exec(compile(module_ast, filename = path, mode = "exec"), globals_defs)
    finally:
        sys.path.pop()

    if (as_dict):
        return globals_defs

    return types.SimpleNamespace(**globals_defs)

def sanitize_code(source_code):
    module_ast = ast.parse(source_code)

    keep_nodes = []
    for node in module_ast.body:
        if (type(node) in AST_NODE_WHITELIST):
            keep_nodes.append(node)
            continue

        if (not isinstance(node, ast.Assign)):
            continue

        if ((len(node.targets) != 1) or (not isinstance(node.targets[0], ast.Name))):
            continue

        if (node.targets[0].id != node.targets[0].id.upper()):
            continue

        keep_nodes.append(node)

    module_ast.body = keep_nodes
    return module_ast

def ast_to_source(module_ast):
    return ast.unparse(module_ast)



import sys

import autograder.api.submission.submit
import autograder.assignment

def run(arguments):
    result = autograder.api.submission.submit.send(arguments, arguments.files, exit_on_error = True)

    if (not result['grading-success']):
        print("Grading failed.")
        return 1

    submission = autograder.assignment.GradedAssignment.from_dict(result['result'])
    print(submission.report())
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.submission.submit._get_parser()

    parser.add_argument('files', metavar = 'FILE',
        action = 'store', type = str, nargs = '+',
        help = 'The path to your submission file(s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



"""
A single question (test case) for an assignment.
"""

import abc
import functools
import json
import numbers
import traceback

import autograder.utils

DEFAULT_TIMEOUT_SEC = 60

class Question(object):
    """
    Questions are grade-able portions of an assignment.
    They can also be thought of as "test cases".
    Note that all scoring is in ints.
    """

    def __init__(self, max_points = 0, name = None, timeout = DEFAULT_TIMEOUT_SEC):
        self.name = name
        if (self.name is None):
            self.name = type(self).__name__

        if ((not isinstance(max_points, numbers.Real)) or (max_points < 0)):
            raise ValueError("max_points must be a real number, got '%s' (type: %s)." % (
                max_points, type(max_points)))

        self.max_points = max_points
        self._timeout = timeout

        # Scoring artifact.
        self.result = GradedQuestion(name = self.name, max_points = self.max_points)

    @abc.abstractmethod
    def score_question(self, submission, **kwargs):
        """
        Assign an actual score to this question.
        The implementer has full access to instance variables.
        However, users should generally just call the grading methods to manipulate the result.
        """

        pass

    def grade(self, submission, additional_data = {}, show_exceptions = False):
        """
        Invoke the scoring method using a timeout and cleanup.
        Return the graded question.
        """

        helper = functools.partial(self._score_helper, submission,
                additional_data = additional_data)

        self._internal_grade(helper, show_exceptions)

        return self.result

    def _internal_grade(self, helper, show_exceptions):
        try:
            success, value = autograder.utils.invoke_with_timeout(self._timeout, helper)
        except Exception:
            if (show_exceptions):
                traceback.print_exc()

            self.set_result(0, "Raised an exception: " + traceback.format_exc())
            return

        if (not success):
            if (value is None):
                self.set_result(0, "Timeout (%d seconds)." % (self._timeout))
            else:
                self.set_result(0, "Error during execution: " + value)

            return

        # Because we use the helper method, we can only get None back if there was an error.
        if (value is None):
            self.set_result(0, "Error running scoring.")
            return

        self.result = value

    def _score_helper(self, submission, additional_data = {}):
        """
        Score the question, but make sure to return the result so
        multiprocessing can properly pass it back.
        """

        self.result = GradedQuestion(name = self.name, max_points = self.max_points)

        self.result.grading_start_time = autograder.utils.get_timestamp()

        try:
            self.score_question(submission, **additional_data)
        except AutograderFailError:
            # The question has been failed, no additional output is required.
            pass

        self.result.grading_end_time = autograder.utils.get_timestamp()

        return self.result

    def get_last_result(self):
        return self.result

    def get_score(self):
        return self.result.score

    # Grading functions.

    def check_not_implemented(self, value):
        if (value is None):
            self.fail("None returned.")

        if (isinstance(value, type(NotImplemented))):
            self.fail("NotImplemented returned.")

        return False

    def set_result(self, score, message):
        self.result.score = score
        self.result.message = message

    def set_score(self, score):
        self.result.score = score

    def set_message(self, message):
        self.result.message = message

    def fail(self, message):
        """
        Immediately fail this question, no partial credit.
        """

        self.set_result(0, message)
        raise AutograderFailError()

    def full_credit(self, message = ''):
        self.set_score(self.max_points)

        if (message != ''):
            self.set_message(message)

    def add_score(self, add_score):
        self.result.score += add_score

    def add_message(self, message, add_score = 0):
        if (self.result.message != ''):
            self.result.message += "\n"

        self.result.message += str(message)
        self.result.score += add_score

    def cap_score(self):
        """
        Cap the score so it is in [0, self.max_points].
        """

        self.result.score = max(0, min(self.max_points, self.result.score))

class AutograderFailError(RuntimeError):
    """
    This error indicates that fail() has been called on a question
    and execution should be stopped.
    """

    pass

class GradedQuestion(object):
    """
    The result of a question being graded with a submission.
    """

    def __init__(self, name = '', max_points = 0,
            score = 0, message = '',
            grading_start_time = None, grading_end_time = None,
            **kwargs):
        self.name = name
        self.max_points = max_points

        self.score = score
        self.message = message

        self.grading_start_time = None
        if (grading_start_time is not None):
            self.grading_start_time = autograder.utils.get_timestamp(grading_start_time)

        self.grading_end_time = None
        if (grading_end_time is not None):
            self.grading_end_time = autograder.utils.get_timestamp(grading_end_time)

    def to_dict(self):
        """
        Convert to all simple structures that can be later converted to JSON.
        """

        return {
            'name': self.name,
            'max_points': self.max_points,
            'score': self.score,
            'message': self.message,
            'grading_start_time': autograder.utils.timestamp_to_string(self.grading_start_time),
            'grading_end_time': autograder.utils.timestamp_to_string(self.grading_end_time),
        }

    @staticmethod
    def from_dict(data):
        """
        Partner to to_dict().
        """

        return GradedQuestion(**data)

    def scoring_report(self, prefix = ''):
        """
        Get a string that represents the scoring for this question.
        """

        if ((prefix != '') and (not prefix.endswith(' '))):
            prefix += ' '

        lines = ["%s%s: %d / %d" % (prefix, self.name, self.score, self.max_points)]
        if (self.message != ''):
            for line in self.message.split("\n"):
                lines.append(prefix + '   ' + line)

        return "\n".join(lines)

    def string(self, indent = None):
        return json.dumps(self.to_dict(), indent = indent)

    def __repr__(self):
        return self.string()

    def __eq__(self, other):
        return self.equals(other)

    def equals(self, other, ignore_messages = False, **kwargs):
        if (not isinstance(other, GradedQuestion)):
            return False

        if (
                (self.name != other.name)
                or (self.max_points != other.max_points)
                or (self.score != other.score)):
            return False

        if (ignore_messages):
            return True

        return self.message == other.message



import sys

import autograder.api.user.change

def run(arguments):
    result = autograder.api.user.change.send(arguments, exit_on_error = True)

    if (not result['found-user']):
        print("User not found.")
        return 0

    print("Password changed.")
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.user.change._get_parser()
    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import autograder.api.common
import autograder.api.peek
import autograder.submission

HEADERS = ['key', 'value', 'source']

def run(arguments):
    config_data, sources = autograder.api.common.parse_config(arguments, show_sources = True)

    print("\t".join(HEADERS))

    for key in sorted(config_data.keys()):
        value = config_data[key]
        if (key == 'pass'):
            value = '*** password omitted ***'

        row = [key, value, sources[key]]
        print("\t".join([str(value) for value in row]))

    return 0

def _get_parser():
    parser = autograder.api.common.get_argument_parser(description =
            'Show all the config values and where those values came from.')

    return parser

def main():
    return run(_get_parser().parse_args())

if (__name__ == '__main__'):
    sys.exit(main())



import contextlib
import os
import sys

from flake8.api import legacy as flake8

import autograder.code
import autograder.question
import autograder.utils

# For codes, see:
# flake8: https://flake8.pycqa.org/en/latest/user/error-codes.html
# pycodestyle: https://pycodestyle.pycqa.org/en/latest/intro.html#error-codes
STYLE_OPTIONS = {
    'max_line_length': 100,
    'max_doc_length': 100,
    'select': 'E,W,F',
    'show_source': True,
    'color': 'never',
    'ignore': [
        # Don't force continuation line alignment.
        'E128',

        # Allow spaces around parameter/keyword equals.
        'E251',

        # Don't force two spaces between functions/class.
        # We only want one space.
        'E302',
        'E305',

        # Allow lambdas to be assigned into a local variable.
        'E731',

        # Ignore missing newlines at the end of the file.
        # This is a result of striping incoming code.
        'W292',

        # Do not enforce spaces in blank lines.
        # This should typically be enforced,
        # but because code from notebooks is hard to give a nice line number for
        # students have difficulty finding where the issue is.
        'W293',

        # PEP-8 recommends breaking a line before a binary operator.
        # This was a recent reversal of idiomatic Python.
        # W503 enforces the old style, while W504 enforces the new style.
        'W503',
    ]
}

class Style(autograder.question.Question):
    """
    A question that can be added to assignments that checks style.
    """

    def __init__(self, paths, max_points = 5, fake_path = None, shorten_path = True):
        super().__init__(max_points)

        if (isinstance(paths, str)):
            paths = [paths]

        if (not isinstance(paths, list)):
            # Allow this to throw.
            paths = list(paths)

        self._paths = paths
        self._fake_path = fake_path
        self._shorten_paths = shorten_path

    def score_question(self, *args, **kwargs):
        error_count, style_output = check_paths(self._paths,
                fake_path = self._fake_path,
                shorten_path = self._shorten_paths)

        if (error_count == 0):
            self.full_credit(message = 'Style is clean!')
            return

        self.add_message(("Code has %d style issues (shown below)."
                + " Note that line numbers may be offset in iPython notebooks.")
                % (error_count))
        self.set_score(max(0, self.max_points - error_count))

        self.add_message("--- Style Output BEGIN ---")
        for (path, lines) in style_output:
            self.add_message("\nStyle Issues for: '%s':" % path)
            self.add_message('---')
            self.add_message("\n".join(lines))
            self.add_message("---\n")
        self.add_message("--- Style Output END ---")

def check_path(path, **kwargs):
    return check_paths([path], **kwargs)

def check_paths(paths, **kwargs):
    """
    Check the style of all the listed paths (recursivley) and return a two-item tuple of:
        - The total number of style violations.
        - A list of two-item tules of:
            - The path to a file with style violations.
            - A list of strings that describe the style issues.
    """

    total_count = 0
    # [(path, lines), ...]
    total_lines = []

    for path in paths:
        if (os.path.isfile(path)):
            if (os.path.splitext(path)[1] not in autograder.code.ALLOWED_EXTENSIONS):
                continue

            count, lines = _check_file(path, **kwargs)
            lines = [(path, lines)]
        else:
            dirents = [os.path.join(path, dirent) for dirent in os.listdir(path)]
            count, lines = check_paths(dirents, **kwargs)

        if (count > 0):
            total_count += count
            total_lines += lines

    return total_count, total_lines

def _check_file(path, fake_path = None, shorten_path = False):
    """
    Check the style of a file and return a two-item tuple of:
        - The number of style violations.
        - A list of strings that describe the style issues.
    """

    if (not os.path.isfile(path)):
        raise ValueError("Can only check style on a file, got a directory: '%s'." % (path))

    cleanup_paths = []

    if (path.endswith('.py')):
        pass
    elif (path.endswith('.ipynb')):
        contents = autograder.code.extract_notebook_code(path)

        temp_path = autograder.utils.get_temp_path(prefix = 'style_', suffix = '_notebook')
        cleanup_paths.append(temp_path)
        with open(temp_path, 'w') as file:
            file.write(contents)

        path = temp_path
    else:
        raise ValueError("Can only check style on .py or .ipynb files, got '%s'." % (path))

    path = os.path.realpath(path)

    replacement_path = path

    if (fake_path is not None):
        replacement_path = fake_path

    if (shorten_path):
        replacement_path = os.path.basename(replacement_path)

    output_path = autograder.utils.get_temp_path(prefix = 'style_', suffix = '_output')
    cleanup_paths.append(output_path)

    # argparse (used by flake8) will look for a program name on sys.argv[0].
    if (len(sys.argv) == 0):
        sys.argv = ['']

    style_guide = flake8.get_style_guide(**STYLE_OPTIONS)

    with open(output_path, 'w') as file:
        with contextlib.redirect_stdout(file):
            report = style_guide.check_files([path])

    with open(output_path, 'r') as file:
        lines = file.readlines()

    lines = [line.rstrip() for line in lines]

    if (path != replacement_path):
        lines = [line.replace(path, replacement_path) for line in lines]

    for path in cleanup_paths:
        autograder.utils.remove_dirent(path)

    return (report._application.result_count, lines)



import atexit
import datetime
import multiprocessing
import os
import re
import shutil
import sys
import tempfile
import time
import traceback
import types
import uuid

import autograder.code

REAP_TIME_SEC = 5

ALL_SUBMISSION_KEY = '__all__'

PRETTY_TIMESTEMP_FORMAT = '%Y-%m-%d %H:%M'

class Mock(object):
    def __init__(self):
        self.item_history = list()
        self.attribute_history = list()
        self.call_history = list()

    def __repr__(self):
        return "Mock -- Item History: %s, Attribute History: %s, Call History: %s" % (
            str(self.item_history), str(self.attribute_history), str(self.call_history))

    def __call__(self, *args, **kwargs):
        self.call_history.append((args, kwargs))
        return self

    def __getitem__(self, name):
        self.item_history.append(name)
        return self

    def __getattr__(self, name):
        self.attribute_history.append(name)
        return self

def _invoke_helper(result, function):
    value = None
    error = None

    try:
        value = function()
    except Exception as ex:
        error = (ex, traceback.format_exc())

    sys.stdout.flush()

    result.put((value, error))
    result.close()

# Return: (success, function return value)
# On timeout, success will be false and the value will be None.
# On error, success will be false and value will be the string stacktrace.
# On successful completion, success will be true and value may be None (if nothing was returned).
def invoke_with_timeout(timeout, function):
    if (not sys.platform.startswith('linux')):
        # Mac and Windows have some pickling issues with multiprocessing.
        # Just run them without a timeout.
        # Any autograder will be run on a Linux machine and will be safe.
        start_time = time.time()
        value = function()
        runtime = time.time() - start_time

        if (runtime > timeout):
            return (False, None)

        return (True, value)

    result = multiprocessing.Queue(1)

    # Note that we use processes instead of threads so they can be more completely killed.
    process = multiprocessing.Process(target = _invoke_helper, args = (result, function))
    process.start()

    # Wait for at most the timeout.
    process.join(timeout)

    # Check to see if the process is still running.
    if (process.is_alive()):
        # Kill the long-running process.
        process.terminate()

        # Try to reap the process once before just giving up on it.
        process.join(REAP_TIME_SEC)

        return (False, None)

    # Check to see if the process explicitly existed (like via sys.exit()).
    if (result.empty()):
        return (False, 'Code explicitly exited (like via sys.exit()).')

    value, error = result.get()

    if (error is not None):
        exception, stacktrace = error
        return (False, stacktrace)

    return (True, value)

def prepare_submission(path, raise_on_collision = False):
    """
    Get a submission from a path, prepare it for grading,
    and return a submission namespace that contains all parsed entities.

    Directories will be fully recursively descended and each python or notebook will be prepared.

    Entries that begin with two underscores will not be included.

    The following things will be done to each file:
        1) The code will be parsed and sanitized.
        2) The sanitized code will be imported as a module into a namespace.
        3) The namespace will be made available in the returned namespace
            in a structure that matches its package structure.
            E.g. "./a/b/c.py" will be available from `prepare_submission('.').a.b.c`,
            and "foo.py" will be available from `prepare_submission('foo.py').foo`.
        4) All entries in the module will be put in the ALL_SUBMISSION_KEY attribute of the
            returned namespace.
            If raise_on_collision is True, an error will be raised if a key already exists.
    """

    submission = {}

    if (os.path.isfile(path)):
        _prepare_submission_file(submission, path, [], raise_on_collision)
    else:
        _prepare_submission_dir(submission, path, [], raise_on_collision)

    return _dict_to_namespace(submission)

def _prepare_submission_dir(submission, path, prefix, raise_on_collision):
    if (not os.path.isdir(path)):
        raise ValueError("Preparation target must be a dir: '%s'." % path)

    for dirent in os.listdir(path):
        dirent_path = os.path.join(path, dirent)

        if (os.path.isfile(dirent_path)):
            if (os.path.splitext(dirent)[1] not in autograder.code.ALLOWED_EXTENSIONS):
                continue

            _prepare_submission_file(submission, dirent_path, prefix, raise_on_collision)
        else:
            _prepare_submission_dir(submission, dirent_path, prefix + [dirent], raise_on_collision)

def _prepare_submission_file(submission, path, prefix, raise_on_collision):
    if (not os.path.isfile(path)):
        raise ValueError("Preparation target must be a file: '%s'." % path)

    basename = os.path.splitext(os.path.basename(path))[0]

    defs = autograder.code.sanitize_and_import_path(path, as_dict = True)

    for (name, value) in defs.items():
        if (name.startswith('__')):
            continue

        if ((name in submission) and (raise_on_collision)):
            raise ValueError("Name collision ('%s') when importing all keys for '%s'." % (
                name, path))

        if (ALL_SUBMISSION_KEY not in submission):
            submission[ALL_SUBMISSION_KEY] = {}

        submission[ALL_SUBMISSION_KEY][name] = defs[name]

    # Place the defs into a structure that matches the path.
    context = submission
    for part in prefix + [basename]:
        if ((part in context) and (not isinstance(context[part], dict))):
            raise ValueError("Name collision ('%s') when importing all keys for '%s'." % (
                part, path))

        if (part not in context):
            context[part] = {}

        context = context[part]

    context.update(defs)

def _dict_to_namespace(root):
    if (not isinstance(root, dict)):
        return root

    for (key, value) in root.items():
        root[key] = _dict_to_namespace(value)

    return types.SimpleNamespace(**root)

def get_temp_path(prefix = '', suffix = '', rm = True):
    """
    Get a path to a valid (but not currently existing) temp dirent.
    If rm is True, then the dirent will be attempted to be deleted on exit
    (no error will occur if the path is not there).
    """

    path = None
    while ((path is None) or os.path.exists(path)):
        path = os.path.join(tempfile.gettempdir(), prefix + str(uuid.uuid4()) + suffix)

    if (rm):
        atexit.register(remove_dirent, path)

    return path

def remove_dirent(path):
    if (not os.path.exists(path)):
        return

    if (os.path.isfile(path) or os.path.islink(path)):
        os.remove(path)
    elif (os.path.isdir(path)):
        shutil.rmtree(path)
    else:
        raise ValueError("Unknown type of dirent: '%s'." % (path))

def get_timestamp(source = None):
    if (source is None):
        return datetime.datetime.now(datetime.timezone.utc)

    if (isinstance(source, datetime.datetime)):
        return source

    if (isinstance(source, str)):
        # Parse out some cases that Python 3.10 cannot deal with.
        # This will remove fractional seconds.
        source = re.sub(r'Z$', '+00:00', source)
        source = re.sub(r'(\d\d:\d\d)(\.\d+)', r'\1', source)

        return datetime.datetime.fromisoformat(source)

    raise ValueError("Unknown type ('%s') for timestamp source." % (type(source)))

def timestamp_to_string(timestamp, pretty = False):
    if (timestamp is None):
        return None

    if (pretty):
        return timestamp.astimezone().strftime(PRETTY_TIMESTEMP_FORMAT)

    return timestamp.isoformat()

def copy_dirent(source, dest):
    """
    Copy a file or directory into dest.
    If source is a file, then dest can be a file or dir.
    If source is a dir, then dest must be a non-existent dir.
    """

    if (os.path.isfile(source)):
        shutil.copy2(source, dest)
    else:
        shutil.copytree(source, dest)

def copy_dirent_contents(source, dest):
    """
    Copy a file or the contents of a directory (excluding the top-level directory) into dest.
    For a file: `cp source dest/`
    For a dir: `cp -r source/* dest/`
    """

    source = os.path.abspath(source)

    if (os.path.isfile(source)):
        shutil.copy2(source, dest)
        return

    for dirent in os.listdir(source):
        source_path = os.path.join(source, dirent)
        dest_path = os.path.join(dest, dirent)

        if (os.path.isfile(source_path)):
            shutil.copy2(source_path, dest_path)
        else:
            shutil.copytree(source_path, dest_path)



import unittest

import autograder.utils

class TestDate(unittest.TestCase):
    def test_timestamp_conversions(self):
        cases = [
            ('2023-09-28', '2023-09-28T00:00:00'),
            ('2023-09-28T00:00:00', '2023-09-28T00:00:00'),
            ('2023-09-28T04:00:20.683684Z', '2023-09-28T04:00:20+00:00'),
            ('2023-09-28T04:00:20.683684+00:00', '2023-09-28T04:00:20+00:00'),
            ('2023-09-28T13:10:44.432050+00:00', '2023-09-28T13:10:44+00:00'),
            ('2023-09-28T13:10:44.43205+00:00', '2023-09-28T13:10:44+00:00'),
        ]

        for (source, expected_string) in cases:
            first_timestamp = autograder.utils.get_timestamp(source)
            string = autograder.utils.timestamp_to_string(first_timestamp)
            second_timestamp = autograder.utils.get_timestamp(string)

            if (expected_string is not None):
                self.assertEquals(string, expected_string)

            self.assertEquals(first_timestamp, second_timestamp)



import sys

import autograder.api.user.get
import autograder.cli.common

def run(arguments):
    result = autograder.api.user.get.send(arguments, exit_on_error = True)

    if (not result['found-user']):
        if (not arguments.table):
            print("No matching user found.")

        return 0

    autograder.cli.common.list_users([result['user']], table = arguments.table)
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.user.get._get_parser()

    parser.add_argument('--table', dest = 'table',
        action = 'store_true', default = False,
        help = 'Output the results as a TSV table with a header (default: %(default)s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import argparse
import json
import os
import sys

import autograder.submission

def run(args):
    grader_path = os.path.abspath(args.grader)
    grading_dir = os.path.abspath(args.dir)

    result = autograder.submission.run_submission(grading_dir, grader_path = grader_path)
    if (result is None):
        return 1

    print(result.report())

    if (args.outpath is not None):
        out_path = os.path.abspath(args.outpath)
        os.makedirs(os.path.dirname(os.path.abspath(out_path)), exist_ok = True)
        with open(out_path, 'w') as file:
            json.dump(result.to_dict(), file, indent = 4)

    return 0

def _get_parser():
    parser = argparse.ArgumentParser(description =
        'Grade a submission given an already prepared grading directory'
        + ' (see autograder.cli.testing.setup-grading-dir)'
        + ' and a grader file.'
        + ' Use autograder.cli.grading.grade if you have not already prepared'
        + ' your grading directory.')

    parser.add_argument('-g', '--grader',
        action = 'store', type = str, required = True,
        help = 'The path to a Python file containing a single assignment class.')

    parser.add_argument('-d', '--dir',
        action = 'store', type = str, required = True,
        help = 'The path to the grading directory.')

    parser.add_argument('-o', '--outpath',
        action = 'store', type = str, required = False, default = None,
        help = 'The path to a output the JSON result.')

    return parser

def main():
    return run(_get_parser().parse_args())

if (__name__ == '__main__'):
    sys.exit(main())



import argparse
import os
import sys

import autograder.submission

def run(args):
    assignment_config_path = os.path.abspath(args.assignment)
    submission_path = os.path.abspath(args.submission)

    grading_dir = autograder.submission.prep_grading_dir(
        assignment_config_path, submission_path,
        grading_dir = args.out_dir,
        skip_static = args.skip_static,
        debug = True)

    print("Prepared grading directory: '%s'." % (grading_dir))

    return 0

def _get_parser():
    parser = argparse.ArgumentParser(description =
        'Setup a directory as if it is being graded.'
        + ' This is useful for seeing what the autograder will see right before grading begins.')

    parser.add_argument('-a', '--assignment',
        action = 'store', type = str, required = True,
        help = 'The path to an assignment JSON file.')

    parser.add_argument('-s', '--submission',
        action = 'store', type = str, required = True,
        help = 'The path to an input submission.')

    parser.add_argument('-o', '--out-dir', dest = 'out_dir',
        action = 'store', type = str, default = None,
        help = 'The base output directory. Will be a temp directory if not specified.')

    parser.add_argument('-ss', '--skip-static', dest = 'skip_static',
        action = 'store_true', default = False,
        help = 'Skip the static portion of the setup (copy and ops) (default: %(default)s)')

    return parser

def main():
    return run(_get_parser().parse_args())

if (__name__ == '__main__'):
    sys.exit(main())



"""
What is necessary to grade a single assignment.
"""

import inspect
import json

import autograder.code
import autograder.question
import autograder.utils

class Assignment(object):
    """
    A collection of questions to be scored.
    """

    def __init__(self,
            name = None,
            questions = [],
            input_dir = '.', output_dir = '.', work_dir = '.',
            prep_submission = True,
            additional_data = {},
            **kwargs):
        """
        Construct an assignment.

        If prep_submission is True, then the default _prepare_submission()
        implementation will prepare the input directory.
        """

        self._name = name
        if (self._name is None):
            self._name = type(self).__name__

        self._questions = questions

        self._input_dir = input_dir
        self._output_dir = output_dir
        self._work_dir = work_dir

        self._prep_submission = prep_submission

        self._additional_data = additional_data

        # Scoring artifact.
        self.result = None

    def grade(self, **kwargs):
        return self._grade_submission(self._prepare_submission(), **kwargs)

    def _grade_submission(self, submission, show_exceptions = False, **kwargs):
        """
        Grade an assignment by grading all the questions.

        The submission argument is the result of _prepare_submission().
        """

        self.result = GradedAssignment(name = self._name, questions = [])
        self.result.grading_start_time = autograder.utils.get_timestamp()

        for question in self._questions:
            self.result.questions.append(question.grade(submission,
                additional_data = self._additional_data,
                show_exceptions = show_exceptions))

        self.result.grading_end_time = autograder.utils.get_timestamp()

        return self.result

    def _prepare_submission(self):
        """
        Prepare the submission in the input directory for grading.
        The result of this is what will be passed to grade().
        Child classes may leave this default behavior or override.

        This implementation will check the prep_submission argument passed in the constructor.
        If true the input directory will be prepared, otherwise None will be returned.
        """

        if (self._prep_submission):
            return autograder.utils.prepare_submission(self._input_dir)

        return None

class GradedAssignment(object):
    """
    The result of an assignment being graded with a submission.
    """

    def __init__(self, name = '',
            questions = [],
            grading_start_time = None, grading_end_time = None,
            **kwargs):
        self.name = name
        self.questions = questions

        self.grading_start_time = None
        if (grading_start_time is not None):
            self.grading_start_time = autograder.utils.get_timestamp(grading_start_time)

        self.grading_end_time = None
        if (grading_end_time is not None):
            self.grading_end_time = autograder.utils.get_timestamp(grading_end_time)

    def to_dict(self):
        """
        Convert to all simple structures that can be later converted to JSON.
        """

        return {
            'name': self.name,
            'questions': [question.to_dict() for question in self.questions],
            'grading_start_time': autograder.utils.timestamp_to_string(self.grading_start_time),
            'grading_end_time': autograder.utils.timestamp_to_string(self.grading_end_time),
        }

    def to_test_submission(self, options = {}):
        """
        Output a dict that can be used as a test submission.
        """

        results = self.to_dict()

        del results['grading_start_time']
        del results['grading_end_time']

        for question in results['questions']:
            del question['grading_start_time']
            del question['grading_end_time']

        test_submission = {
            'result': results,
        }

        test_submission.update(options)

        return test_submission

    @staticmethod
    def from_dict(data):
        """
        Partner to to_dict().
        """

        data = dict(data)

        if ('questions' in data):
            data['questions'] = [
                autograder.question.GradedQuestion(**question)
                for question in data['questions']
            ]

        return GradedAssignment(**data)

    def get_score(self):
        """
        Return (total score, max score).
        """

        total_score = 0
        max_score = 0

        for question in self.questions:
            total_score += question.score
            max_score += question.max_points

        return (total_score, max_score)

    def report(self, prefix = ''):
        """
        Return a string representation of the grading for this assignment.
        """

        if ((prefix != '') and (not prefix.endswith(' '))):
            prefix += ' '

        output = [
            prefix + "Autograder transcript for assignment: %s." % (self.name),
            prefix + "Grading started at %s and ended at %s." % (
                autograder.utils.timestamp_to_string(self.grading_start_time, pretty = True),
                autograder.utils.timestamp_to_string(self.grading_end_time, pretty = True))
        ]

        total_score = 0
        max_score = 0

        for question in self.questions:
            total_score += question.score
            max_score += question.max_points

            output.append(question.scoring_report(prefix = prefix))

        output.append('')
        output.append(prefix + "Total: %d / %d" % (total_score, max_score))

        return "\n".join(output)

    def __eq__(self, other):
        return self.equals(other)

    def string(self, indent = None):
        return json.dumps(self.to_dict(), indent = indent)

    def __repr__(self):
        return self.string()

    def equals(self, other, **kwargs):
        if (not isinstance(other, GradedAssignment)):
            return False

        if ((self.name != other.name) or (len(self.questions) != len(other.questions))):
            return False

        for i in range(len(self.questions)):
            if (not self.questions[i].equals(other.questions[i], **kwargs)):
                return False

        return True

def load_assignments(path):
    """
    Recursively load all the assignments in a path (file or dir).
    """

    module = autograder.code.sanitize_and_import_path(path)
    assignments = []

    for name in dir(module):
        obj = getattr(module, name)

        if (not inspect.isclass(obj)):
            continue

        if (obj == autograder.assignment.Assignment):
            continue

        if (issubclass(obj, (autograder.assignment.Assignment,))):
            assignments.append(obj)

    return assignments

def fetch_assignment(path):
    """
    Recursively fetch a single assignment from a path.
    If exactly one assignment is not found, then raise an error.
    """

    assignments = load_assignments(path)

    if (len(assignments) == 0):
        raise ValueError(("Assignment file (%s) does not contain any instances of"
            + " autograder.assignment.Assignment.") % (path))
    elif (len(assignments) > 1):
        raise ValueError(("Assignment file (%s) contains more than one (%d) instances of"
            + " autograder.assignment.Assignment.") % (path, len(assignments)))

    return assignments[0]



import sys

import autograder.api.constants
import autograder.api.user.add
import autograder.cli.common
import autograder.util.hash

def run(arguments):
    arguments = vars(arguments)

    password = arguments['new-pass']
    if (password != ''):
        password = autograder.util.hash.sha256_hex(password)

    arguments['new-users'] = [{
        'email': arguments['new-email'],
        'pass': password,
        'name': arguments['new-name'],
        'role': arguments['new-role'],
        'lms-id': arguments['new-lms-id'],
    }]

    result = autograder.api.user.add.send(arguments, exit_on_error = True)

    autograder.cli.common.list_add_users(result, table = arguments['table'])
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.user.add._get_parser()

    parser.description = ('Add a user to the course.'
            + ' When force is true, this becomes an upsert (update if exists, otherwise insert).')

    parser.add_argument('--new-email', dest = 'new-email',
        action = 'store', type = str, required = True,
        help = 'The email of the user to add.')

    parser.add_argument('--new-pass', dest = 'new-pass',
        action = 'store', type = str, default = '',
        help = 'The password of the user to add.'
            + ' If empty, the server will generate and email a password.')

    parser.add_argument('--new-name', dest = 'new-name',
        action = 'store', type = str, default = '',
        help = 'The name of the user to add.')

    parser.add_argument('--new-role', dest = 'new-role',
        action = 'store', type = str, default = 'unknown',
        choices = autograder.api.constants.ROLES,
        help = 'The role of the user to add (defaults to student).')

    parser.add_argument('--new-lms-id', dest = 'new-lms-id',
        action = 'store', type = str, default = '',
        help = 'Role of the user to add (defaults to student).')

    parser.add_argument('--table', dest = 'table',
        action = 'store_true', default = False,
        help = 'Output the results as a TSV table with a header (default: %(default)s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import json
import os

import autograder.utils
import autograder.util.file
import autograder.util.gzip

def output_grading_result(result, base_dir = '.'):
    out_dir = os.path.join(base_dir, result['info']['id'])
    if (os.path.exists(out_dir)):
        autograder.utils.remove_dirent(out_dir)

    os.makedirs(out_dir, exist_ok = True)

    stdout_path = os.path.join(out_dir, 'stdout.txt')
    autograder.util.file.write(stdout_path, result['stdout'])

    stderr_path = os.path.join(out_dir, 'stderr.txt')
    autograder.util.file.write(stderr_path, result['stderr'])

    result_path = os.path.join(out_dir, 'info.json')
    autograder.util.file.write(result_path, json.dumps(result['info'], indent = 4))

    grading_input_dir = os.path.join(out_dir, 'input')
    _output_grading_result_dir(grading_input_dir, result['input-files-gzip'])

    grading_output_dir = os.path.join(out_dir, 'output')
    _output_grading_result_dir(grading_output_dir, result['output-files-gzip'])

def _output_grading_result_dir(out_dir, files):
    os.makedirs(out_dir, exist_ok = True)

    for (relpath, gzip_contents) in files.items():
        path = os.path.join(out_dir, *relpath.split('/'))
        os.makedirs(os.path.dirname(path), exist_ok = True)

        contents = autograder.util.gzip.from_base64(gzip_contents)
        autograder.util.file.write(path, contents)



def read(path, strip = True):
    with open(path, 'r') as file:
        contents = file.read()

    if (strip):
        contents = contents.strip()

    return contents

def write(path, contents, strip = True, newline = True):
    if (contents is None):
        contents = ''

    if (strip):
        contents = contents.strip()

    if (newline):
        contents += "\n"

    with open(path, 'w') as file:
        file.write(contents)



import sys

import autograder.api.lms.syncusers
import autograder.cli.common

def run(arguments):
    result = autograder.api.lms.syncusers.send(arguments, exit_on_error = True)
    autograder.cli.common.list_sync_users(result, table = arguments.table)
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.lms.syncusers._get_parser()

    parser.add_argument('--table', dest = 'table',
        action = 'store_true', default = False,
        help = 'Output the results as a TSV table with a header (default: %(default)s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())
